/*
 * transport.c 
 *
 *	Project 3		
 *
 * This file implements the STCP layer that sits between the
 * mysocket and network layers. You are required to fill in the STCP
 * functionality in this file. 
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <arpa/inet.h>
#include "mysock.h"
#include "stcp_api.h"
#include "transport.h"
#include <sys/time.h>

#define MAX_SEQ_NUM 256
#define WINDOW_SIZE 3072
#define MAX_SEND 6
#define OFFSET 5
#define TIME_OUT 2

#define DEBUG 0

enum { 
    CSTATE_WAIT_FOR_ACK,
    CSTATE_SEND_FIN,
    CSTATE_FIN_RECVD,
    CSTATE_WAIT_FOR_FIN,

    CSTATE_CLOSED,
    CSTATE_ESTABLISHED
};    /* you should have more states */

/* this structure is global to a mysocket descriptor */
typedef struct
{
    bool_t done;    /* TRUE once connection is closed */

    int connection_state;   /* state of the connection (established, etc.) */
    tcp_seq initial_sequence_num;

    char sendBuff[WINDOW_SIZE];     /* The send buffer */
    char recBuff[WINDOW_SIZE];      /* Receive buffer */

    tcp_seq nextSeqSend;   /* Next sequence number to be sent from here */
    tcp_seq lastSeqRecv;     /* Last sequence number received from the other host */
    tcp_seq lastSeqAcked;        /* Last sequence number acked from the other host */
    
    int sendWinSize;
    int recWinSize;
    int advWinSize;     /* Advertised window size by the other one */

    bool closeInit;     /* Stores whether we've initiated the fin sequence */

    /* any other connection-wide global variables go here */
} context_t;


void printContext(context_t *ctx) {
    printf("_______________START_CONTEXT____________________\n");
    printf("nextSeqSend : %u\n", ctx->nextSeqSend);
    printf("lastSeqRecv : %u\n", ctx->lastSeqRecv);
    printf("lastSeqAcked : %u\n", ctx->lastSeqAcked);
    printf("sendWinSize : %d\n", ctx->sendWinSize);
    printf("recWinSize : %d\n", ctx->recWinSize);
    printf("advWinSize : %d\n", ctx->advWinSize);
    printf("closeInit : %d\n", (int) ctx->closeInit);
    printf("_____________FINISH_CONTEXT____________________\n");
}

void printPack(struct tcphdr *pack) {
    printf("_______________START_PACK____________________\n");
    printf("Seq : %u\n", pack->th_seq);
    printf("Ack : %u\n", pack->th_ack);
    printf("Flags : %u\n", pack->th_flags);
    printf("Window : %u\n", pack->th_win);
    printf("_____________FINISH_PACK____________________\n");
}

static void generate_initial_seq_num(context_t *ctx);
static void control_loop(mysocket_t sd, context_t *ctx);
int tearDown(mysocket_t sd, context_t *ctx);

/* initialise the transport layer, and start the main loop, handling
 * any data from the peer or the application.  this function should not
 * return until the connection is closed.
 */
void transport_init(mysocket_t sd, bool_t is_active)
{

    if (DEBUG)
        printf ("Entering func transport_init()\n");

    context_t *ctx;

    ctx = (context_t *) calloc(1, sizeof(context_t));
    assert(ctx);

    generate_initial_seq_num(ctx);
    // Updates the initial seq num field in ctx

    /* XXX: you should send a SYN packet here if is_active, or wait for one
     * to arrive if !is_active.  after the handshake completes, unblock the
     * application with stcp_unblock_application(sd).  you may also use
     * this to communicate an error condition back to the application, e.g.
     * if connection fails; to do so, just set errno appropriately (e.g. to
     * ECONNREFUSED, etc.) before calling the function.
     */

    int sendCount = 0;

    ctx->nextSeqSend = ctx->initial_sequence_num;
    ctx->sendWinSize = ctx->recWinSize = WINDOW_SIZE;
    ctx->advWinSize = 0;
    ctx->lastSeqRecv = 0;
    ctx->lastSeqAcked = 0;
    // Invalid for now

    if (DEBUG)
        printf ("Initial sequence number: %d\n", ctx->initial_sequence_num);

    if (is_active) {

        if (DEBUG)
            printf ("In active mode: Sending SYN packet\n");

        struct tcphdr *synHead;
        synHead = (struct tcphdr*) calloc(1, sizeof(struct tcphdr));

        synHead->th_flags = TH_SYN;
        synHead->th_win = ctx->recWinSize;
        synHead->th_seq = ctx->nextSeqSend;

        sendCount = 0;

        if (DEBUG)
            printPack(synHead);
            // Add debug statement

        while (1) {
            if (sendCount >= MAX_SEND) {
                errno = ECONNREFUSED;
                // Setting errno to the approporiate state
                printf("ERROR: Maximum number of sends exhausted\n");
                return;
            }
            // Sending prepared packet
            stcp_network_send(sd, synHead, sizeof(struct tcphdr), NULL);
            sendCount++;

            if (stcp_wait_for_event(sd, NETWORK_DATA, NULL) == NETWORK_DATA)
                break;
        }

        if (DEBUG)
            printf("SYN Header sent! SeqNum: %d\n", synHead->th_seq);
          
        sendCount = 0;
        struct tcphdr *synAck;
        synAck = (struct tcphdr *) calloc(1, sizeof(struct tcphdr));
        stcp_network_recv(sd, synAck, STCP_MSS);

        if (!((synAck->th_flags) & TH_ACK)) {
            // Check the corresponding bit
            printf("ERROR: Ack bit is not set\n");
            return;
        }

        if (!((synAck->th_flags) & TH_SYN)) {
            printf("ERROR: SYN is not set in packet\n");
            return;
        }
      
        if (DEBUG)
            printf("ACK received for: %d\n", synAck->th_ack);

        ctx->lastSeqRecv = synAck->th_seq;
        ctx->nextSeqSend++;
        ctx->advWinSize = synAck->th_win;
        ctx->sendWinSize = MIN(ctx->advWinSize, WINDOW_SIZE);

        struct tcphdr *ackMsg;
        ackMsg = (struct tcphdr *) calloc(1, sizeof(struct tcphdr));
        
        ackMsg->th_seq = ctx->nextSeqSend;
        ackMsg->th_flags = TH_ACK;
        ackMsg->th_win = WINDOW_SIZE;
        ackMsg->th_ack = ctx->lastSeqRecv + 1;

        stcp_network_send(sd, ackMsg, sizeof(struct tcphdr), NULL);

        if (DEBUG)
            printf("ACK sent! SeqNum: %d, Ack for: %d\n", ackMsg->th_seq, ackMsg->th_ack);

        if (synHead)
            free(synHead);

        if (synAck)
            free(synAck);

        if (ackMsg)
            free(ackMsg);

    }
    else {

        if (DEBUG)
            printf ("In passive mode: Waiting for SYN packet\n");

        stcp_wait_for_event(sd, NETWORK_DATA, NULL);
        
        struct tcphdr *synPack;
        synPack = (struct tcphdr *) calloc(1, sizeof(struct tcphdr));
        stcp_network_recv(sd, synPack, STCP_MSS);
        // Note that STCP_MSS has the maximum segment length

        if (!((synPack->th_flags) & TH_SYN)) {
            printf("ERROR: SYN is not set in packet\n");
            return;
        }
        if (DEBUG)
            printf("SYN received! Seq Num : %d\n", synPack->th_seq);

        ctx->advWinSize = synPack->th_win;
        ctx->sendWinSize = MIN(ctx->advWinSize, WINDOW_SIZE);
        ctx->lastSeqRecv = synPack->th_seq;

        struct tcphdr *synAck;
        synAck = (struct tcphdr *) calloc(1, sizeof(struct tcphdr));
        synAck->th_ack = ctx->lastSeqRecv + 1;
        synAck->th_win = WINDOW_SIZE;
        synAck->th_flags = (TH_SYN | TH_ACK);
        synAck->th_seq = ctx->nextSeqSend;

        ctx->nextSeqSend++;

        sendCount = 0;

        while (1) {
            if (sendCount >= MAX_SEND) {
                errno = ECONNREFUSED;
                // Setting errno to the approporiate state
                printf("ERROR: Maximum number of sends exhausted\n");
                return;
            }

            // Sending prepared packet
            stcp_network_send(sd, synAck, sizeof(struct tcphdr), NULL);
            sendCount++;

            if (stcp_wait_for_event(sd, NETWORK_DATA, NULL) == NETWORK_DATA)
                break;

        }

        if (DEBUG)
            printf("SYN ACK sent! SeqNum: %d, Ack for: %d\n", synAck->th_seq, synAck->th_ack);
        
        struct tcphdr * ack;
        ack = (struct tcphdr *) (calloc(1, sizeof(struct tcphdr)));
        stcp_network_recv(sd, ack, STCP_MSS);

        if (!((ack->th_flags) & TH_ACK)) {
            // Check the corresponding bit
            printf("ERROR: Ack bit is not set\n");
            return;
        }

        // if ((ack->th_ack) != (ctx->initial_sequence_num + 1)) {
        //     printf("ERROR: Received out of sequence packet\n");
        //     return;
        // }

        if (DEBUG)
            printf("ACK received! Seq Num: %d, Ack for: %d\n", ack->th_seq, ack->th_ack);

        ctx->advWinSize = ack->th_win;
        ctx->lastSeqRecv = ack->th_seq;
        // ctx->lastSeqAcked = ack->th_seq;

        if (ack)
            free(ack);

        if (synAck)
            free(synAck);

    }

    if (DEBUG)
        printf("Handshake successful: Connection established!\n");

    ctx->connection_state = CSTATE_ESTABLISHED;
    stcp_unblock_application(sd);

    control_loop(sd, ctx);

    /* do any cleanup here */
    if (ctx)
        free(ctx);

}

/* generate random initial sequence number for an STCP connection */
static void generate_initial_seq_num(context_t *ctx)
{
    assert(ctx);

#ifdef FIXED_INITNUM
    /* please don't change this! */
    ctx->initial_sequence_num = 1;
#else
    /* you have to fill this up */
    /* Generate a random number in the range 0-255 */
    ctx->initial_sequence_num = rand() % MAX_SEQ_NUM;
    if (DEBUG)
        printf("Initial sequence number: %d\n", ctx->initial_sequence_num);
#endif
}

// Create a timespec timeout object with the passed delay
timespec getTimeout(int timeSec) {
    timeval currentTime;
    gettimeofday(&currentTime, NULL);
    timespec timeout;
    timeout.tv_sec  = currentTime.tv_sec + timeSec;
    // Add dealy from cuurent
    timeout.tv_nsec = currentTime.tv_usec * 1000;
    // Conversion form nano to micro
    return timeout;
}

/* control_loop() is the main STCP loop; it repeatedly waits for one of the
 * following to happen:
 *   - incoming data from the peer
 *   - new data from the application (via mywrite())
 *   - the socket to be closed (via myclose())
 *   - a timeout
 */
static void control_loop(mysocket_t sd, context_t *ctx)
{
    assert(ctx);
    assert(!ctx->done);

    if (DEBUG)
        printContext(ctx);


    // Reinitalize the ctx variables, for consistency
    ctx->lastSeqAcked = ctx->lastSeqRecv + 1;
    ctx->sendWinSize = MIN(ctx->advWinSize, WINDOW_SIZE);
    ctx->recWinSize = WINDOW_SIZE;

    while (!ctx->done)
    {

        unsigned int eventId;

        timespec timeout = getTimeout(TIME_OUT);

        /* see stcp_api.h or stcp_api.c for details of this function */
        /* XXX: you will need to change some of these arguments! */
        eventId = stcp_wait_for_event(sd, ANY_EVENT, &timeout);

        ctx->sendWinSize = (WINDOW_SIZE - (ctx->nextSeqSend - ctx->lastSeqAcked - 1));

        /* check whether it was the network, app, or a close request */
        if (eventId & APP_DATA)
        {
             
            char *rcvData;
            int rcvSize;

            int availRec = (MIN(ctx->advWinSize, WINDOW_SIZE)) - sizeof(struct tcphdr);
            int availSend = ctx->sendWinSize;

            if (MIN(availSend,availRec) > 0) {

                if (DEBUG) {
                    printf ("App Data mode: App has requested data to be sent\n");
                    // printContext(ctx);
                }

                struct tcphdr *header;
                header = (struct tcphdr *) calloc(1, sizeof(struct tcphdr));
                header->th_off = OFFSET;
                header->th_win = ctx->recWinSize;
                header->th_seq = ctx->nextSeqSend;
                header->th_flags = 0;
                // Forming the header to send

                availRec = MIN(availSend, availRec);
                if (availRec > STCP_MSS)
                    availRec = STCP_MSS;

                rcvData = (char *) malloc(availRec * sizeof(char));
                rcvSize = stcp_app_recv(sd, rcvData, availRec);

                if (DEBUG) {
                    printf("availRec, availSend: %d, %d\n", availRec, availSend);
                }

                while (stcp_network_send(sd, header, sizeof(struct tcphdr), rcvData, rcvSize, NULL) == -1);
                // Repeatedly keep sending

                int startNum = ctx->nextSeqSend % WINDOW_SIZE;
                int endNum = (startNum + rcvSize);

                if (DEBUG)
                    printf("start, end: %d, %d\n", startNum, endNum);

                int i;
                if (endNum < WINDOW_SIZE) {
                    for (i = startNum; i < endNum; i++)
                        ctx->sendBuff[i] = rcvData[i - startNum];
                        // Storing the data at the appropriate place in the buffer
                }
                else {
                    // Wrap around, Extracting the elements in the first part
                    for (i = startNum; i < WINDOW_SIZE; i++)
                        ctx->sendBuff[i] =  rcvData[i - startNum];
                    // Extracting the elements in the second part
                    for (i = 0; i < endNum % WINDOW_SIZE; i++)
                        ctx->sendBuff[i] = rcvData[WINDOW_SIZE - startNum + i];
                }
    
                ctx->nextSeqSend += rcvSize;

                if (DEBUG)
                    printf("Data packet of size: %d, Seq Num: %d sent\n", rcvSize, ctx->nextSeqSend);

                if (rcvData)
                    free(rcvData);

            }
        }

        if (eventId & NETWORK_DATA)
        {

            if (DEBUG) {
                printf ("Network Data mode: Network wants to deliver data\n");
                // printContext(ctx);
            }
            
            // Get segment from network
            int segLen = 2 * STCP_MSS;
            // Upper limit to size
            char *segData;
            segData = (char *) malloc(segLen * sizeof(char));
            segLen = stcp_network_recv(sd, segData, segLen);
            // Update segLen to the actual amount of data

            if (DEBUG)
                printf("Segment accepted from network\n");

            tcphdr *recvMsg = (tcphdr *) segData;
            // Read the initial part of the message
            ctx->advWinSize = recvMsg->th_win;

            if (DEBUG)
                printPack(recvMsg);

            if (recvMsg->th_flags & TH_ACK) {
                // If it is an acknowledgement
                if (DEBUG)
                    printf("ACK message received, seqNum: %d, ack for: %d\n", recvMsg->th_seq, recvMsg->th_ack);
                ctx->lastSeqAcked = recvMsg->th_ack;
            }

            int segSize = segLen - recvMsg->th_off * 4;
            // Getting the part with the data
            ctx->lastSeqRecv = recvMsg->th_seq;

            if (segSize != 0) {
                
                if (DEBUG) {
                    printf("Entering data transfer part. Segment contains data with seqNum: %d\n", recvMsg->th_seq);
                    printf("Current Send Window Size: %d\n", ctx->sendWinSize);
                }

                char *actData = segData + recvMsg->th_off * 4;

                struct tcphdr *ackMsg;
                ackMsg = (struct tcphdr *) (calloc(1, sizeof(struct tcphdr)));
                ackMsg->th_flags = TH_ACK;
                ackMsg->th_off = OFFSET;
                ackMsg->th_win = ctx->recWinSize;
                ackMsg->th_ack = ctx->lastSeqRecv + segSize;
                ackMsg->th_seq = ctx->nextSeqSend;
                // Prepare ack message

                stcp_network_send(sd, ackMsg, sizeof(tcphdr), NULL);
                // Send the ack message

                if (DEBUG) {
                    // print(ackMsg);
                    printf("ACK message sent!: SeqNum: %d, Ack for: %d\n", ackMsg->th_seq, ackMsg->th_ack);
                }

                stcp_app_send(sd, actData, segSize);
                // Sending data to app, note that appData contains the data part

                if (ackMsg)
                    free(ackMsg);

            }
            
            if (recvMsg->th_flags & TH_FIN) {
                if (DEBUG)
                    printf("FIN received: Initiating FIN-ACK sequence\n");
                ctx->connection_state = CSTATE_FIN_RECVD;
                ctx->closeInit = false;
                tearDown(sd, ctx);
            }
        }

        if (eventId & APP_CLOSE_REQUESTED) {
            if (DEBUG)
                printf("APP requests to close connection: Initiating FIN sequence\n");
            ctx->connection_state = CSTATE_SEND_FIN;
            ctx->closeInit = true;
            tearDown(sd, ctx);
        }
    }  
}

int tearDown(mysocket_t sd, context_t *ctx) {

    int count = 0;
    timespec timeout;
    unsigned int stat;

    // Keep trying to close till you pass a threshold
    while (ctx->connection_state != CSTATE_CLOSED)
    {

        if (DEBUG)
            printContext(ctx);

        switch (ctx->connection_state) {

            case CSTATE_SEND_FIN:

                if (count >= MAX_SEND) {
                    errno = ECONNREFUSED;
                    return -1;
                }
                if (DEBUG)
                    printf("Preparing to send FIN\n");

                // Prepare fin header
                struct tcphdr *finHead;
                finHead = (struct tcphdr *) (calloc(1, sizeof(struct tcphdr)));

                finHead->th_flags = TH_FIN;
                finHead->th_seq = ctx->nextSeqSend;
                finHead->th_win = ctx->recWinSize;
                finHead->th_off = OFFSET;

                ctx->nextSeqSend++;

                if (DEBUG)
                    printPack(finHead);

                while (stcp_network_send(sd, finHead, sizeof(struct tcphdr), NULL) == -1);

                if (DEBUG)
                    printf("FIN message sent! seqNum: %d\n", finHead->th_seq);

                ctx->connection_state = CSTATE_WAIT_FOR_ACK;
                // Update connection state

                if (finHead) {
                    free(finHead);
                }

                count++;
                break;

            case CSTATE_WAIT_FOR_ACK:

                if (DEBUG)
                    printf("Currently waiting for ACK\n");

                timeout = getTimeout(TIME_OUT);
                stat = stcp_wait_for_event(sd, NETWORK_DATA, &timeout);

                if (stat & NETWORK_DATA) {

                    // Get segment from network
                    int segLen = 2 * STCP_MSS;
                    // Upper limit to size
                    char *segData;
                    segData = (char *) malloc(segLen * sizeof(char));
                    segLen = stcp_network_recv(sd, segData, segLen);
                    // Update segLen to the actual amount of data

                    if (DEBUG)
                        printf("Segment accepted from network\n");

                    tcphdr *recvMsg = (tcphdr *) segData;

                    if (!(recvMsg->th_flags & TH_ACK) || (recvMsg->th_ack != ctx->nextSeqSend)) {
                        if (DEBUG)
                            printf("FIN ACK not received: Trying again\n");
                        ctx->connection_state = CSTATE_SEND_FIN;
                    }
                    else {
                        if (DEBUG)
                            printf("FIN ACK received, seqNum: %d, ack for: %d\n", recvMsg->th_seq, recvMsg->th_ack);
                        if (ctx->closeInit)
                            // Now need to wait for a fin, since we started it
                            ctx->connection_state = CSTATE_WAIT_FOR_FIN;
                        else
                            // Since we received a fin ack, thus the connection is successfully closed now
                            ctx->connection_state = CSTATE_CLOSED;
                        ctx->lastSeqRecv = recvMsg->th_seq;
                        count = 0;
                    }

                    if (segData)
                        free(segData);

                }
                else
                    ctx->connection_state = CSTATE_SEND_FIN;
                    // Try again

                break;

            case CSTATE_WAIT_FOR_FIN:

                if (DEBUG)
                    printf("Currently waiting for FIN\n");

                timeout = getTimeout(TIME_OUT);
                stat = stcp_wait_for_event(sd, NETWORK_DATA, &timeout);

                if (stat & NETWORK_DATA) {
                    
                    // Get segment from network
                    int segLen = 2 * STCP_MSS;
                    // Upper limit to size
                    char *segData;
                    segData = (char *) malloc(segLen * sizeof(char));
                    segLen = stcp_network_recv(sd, segData, segLen);
                    // Update segLen to the actual amount of data

                    if (DEBUG)
                        printf("Segment accepted from network\n");

                    tcphdr *recvMsg = (tcphdr *) segData;

                    if (!(recvMsg->th_flags & TH_FIN)) {
                        if (DEBUG)
                            printf("FIN not received\n");
                    }
                    else {
                        if (DEBUG)
                            printf("FIN message received, seqNum %d\n", recvMsg->th_seq);
                        ctx->lastSeqRecv = recvMsg->th_seq;
                        ctx->connection_state = CSTATE_FIN_RECVD;
                    }

                    if (segData)
                        free(segData);

                }

                break;

            case CSTATE_FIN_RECVD:

                if (DEBUG)
                    printf("Currently preparing to send ACK\n");

                // Prepare ack header
                struct tcphdr *ackHead;
                ackHead = (struct tcphdr *) (calloc(1, sizeof(struct tcphdr)));    

                ackHead->th_flags = TH_ACK;
                ackHead->th_win = ctx->recWinSize;
                ackHead->th_off = OFFSET;
                ackHead->th_seq = ctx->nextSeqSend;
                ackHead->th_ack = ctx->lastSeqRecv + 1;

                // Send the fin ack
                while (stcp_network_send(sd, ackHead, sizeof(struct tcphdr), NULL) == -1);

                if (ctx->closeInit)
                    ctx->connection_state = CSTATE_CLOSED;
                else
                    ctx->connection_state = CSTATE_SEND_FIN;

                if (DEBUG)
                    printf("ACK message sent!\n");

                if (ackHead)
                    free(ackHead);

                break;

            default:
                break;

        }
    }

    if (DEBUG)
        printf("Connection successfully closed!\n");

    ctx->done = true;

    return 1;
}

/**********************************************************************/
/* our_dprintf
 *
 * Send a formatted message to stdout.
 * 
 * format               A printf-style format string.
 *
 * This function is equivalent to a printf, but may be
 * changed to log errors to a file if desired.
 *
 * Calls to this function are generated by the dprintf amd
 * dperror macros in transport.h
 */
void our_dprintf(const char *format,...)
{
    va_list argptr;
    char buffer[1024];

    assert(format);
    va_start(argptr, format);
    vsnprintf(buffer, sizeof(buffer), format, argptr);
    va_end(argptr);
    fputs(buffer, stdout);
    fflush(stdout);
}



