#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <cstring>
#include <string>
#include <sstream>
#include <vector>
#include <algorithm>
#include <fstream>

#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <sys/wait.h>
#include <fcntl.h>

#include "proxy_parse.h"

using namespace std;

#define MAX_CLIENTS 100
#define MAX_CHILD 20
#define BUF_SIZE 4096
#define DEBUGFLAG 0

typedef struct sockaddr_in socketAddr;
typedef struct sockaddr sockAddr;

string CRLF = "\r\n";

// Safely closes the socket with the provided socket Id

void closeSocket(int socketId) {
	close(socketId);
	shutdown(socketId, SHUT_RDWR);
}

class ProxyServer {

public:

	~ProxyServer() {
		if (socketId >= 0) {
			closeSocket(socketId);
		}
		if (DEBUGFLAG) {
			cout << "ProxyServer closed!\n";
		}
	}

	int start(int portId);
	void run();
	
	int portNum;
	int socketId;
	int childNum;

	int sendRequest(ParsedRequest *reqParse);
	string readRequest(int sockId);
	int handleRequests(int sockId);
	void createConnectionThread(int sockId);
};

// Initialize the proxy server by opening a connection on portNum
// It keeps listening for client connections

int ProxyServer::start(int portId) {

	portNum = -1;
	socketId = -1;
	childNum = 0;

	socketAddr serverAddr;
	portNum = portId;
 
	// Socket Id to which we listen
	// Upon successful completion, socket() shall return a non-negative integer, 
	// the socket file descriptor. Otherwise, a value of -1 shall be returned.
	socketId = socket(AF_INET, SOCK_STREAM, 0);

	if (socketId < 0) {
		printf("Could not retreive socket\n");
		return -1;
	}

	// To reuse the same port
	int reusePort = 1;
	if (setsockopt(socketId, SOL_SOCKET, SO_REUSEADDR, (const char*)&reusePort, sizeof(reusePort)) < 0) {
	    if (DEBUGFLAG) 
	    	cout << "start: Could not set reuse mode\n";
	    return -1;
	}

	if (DEBUGFLAG)
		printf("Succeeded in Socket retrieval\n");

	serverAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	serverAddr.sin_family = AF_INET;
	serverAddr.sin_port = htons(portNum);

	/* Bind the address struct to the socket */
	// socket -> bind -> listen -> accept -> doStuff
	bind(socketId, (sockAddr*) &serverAddr, sizeof(serverAddr));

	// Listen with the max number of connections
	// On success, zero is returned. On error, -1 is returned
	if(listen(socketId, MAX_CLIENTS) < 0) {
		printf("Could not listen\n");
		return -1;
	}

	return 0;
}

// Run the proxy server. It accepts all incoming client connections using 
// an infinite loop, upon a connection request forks threads to handle them

void ProxyServer::run() {

	while (1) {
		int sockId = accept(socketId, (sockAddr*) NULL, NULL);
		// Message denoting new connection established
		if (DEBUGFLAG)
			printf("run: Connection established with client\n");

		if (sockId < 0)
			printf("run: No available slots for the client\n");

	    // Create another thread
	    createConnectionThread(sockId);
	}
}

// handleRequests: Handles the client requests which performs
// a complete proxy transaction by reading the client input, parsing
// it, then sending it to the remote server and receiving and forward
// to the client.

int ProxyServer::handleRequests(int sockId) {
	
	// Get the request from the client
	string request = readRequest(sockId);
	if (request == "") {
		if (DEBUGFLAG)
			cout << "handleRequests: Unable to get request\n";
		return -1;
	}
	string response = "";

	ParsedRequest *reqParse = ParsedRequest_create();
	// Parsing the request
	if (ParsedRequest_parse(reqParse, request.c_str(), request.size()) < 0) {
		if (DEBUGFLAG)
			cout << "handleRequests: Parsing error: Parsing failed for client request\n";
		response = "HTTP/1.0 500 Internal Error" + CRLF + CRLF;
  	}
  	else {
  		// Parsing successful
      	int serverSockId = sendRequest(reqParse);
    	if (serverSockId < 0) {
			if (DEBUGFLAG)
				cout << "handleRequests: Server unable to handle request\n";
	    	response = "HTTP/1.0 500 Internal Error" + CRLF + CRLF;
    	} 
    	else {
    		char buffer[BUF_SIZE + 1];
      		int bufferSize = BUF_SIZE;

      		memset(buffer, '\0', bufferSize);
	        int readNum = 0;
		    // Read response
		    while ((readNum = read(serverSockId, buffer, bufferSize)) > 0) {
		    	write(sockId, buffer, readNum);
		    	memset(buffer, '\0', bufferSize);
		    }
		    if (readNum < 0) {
		        response = "HTTP/1.0 500 Internal Error" + CRLF + CRLF;
				if (DEBUGFLAG)
					cout << "handleRequests: Server closed\n";
		    }
		    closeSocket(serverSockId);
    	}
	}

	if (response.size() > 0) {
		if (DEBUGFLAG)
			cout << "handleRequests: The response is " << response << endl;
		if (write(sockId, response.c_str(), response.size()) <= 0) {
			if (DEBUGFLAG)
				cout << "handleRequests: Unable to write response\n";
		}
	}

	ParsedRequest_destroy(reqParse);

	return 0;
}

// createConnectionThread() creates new threads for incoming client
// connections. It also updates the number of active threads.
// As described in the project specification, it first clears all the
// dead children and in case the number of children are more than the
// maximum allowed number, it waits for the children to die.

void ProxyServer::createConnectionThread(int sockId) {

	while (waitpid(-1, NULL, WNOHANG) > 0) {
		// Wait for all dead children
		childNum--;
	}

	if (childNum >= MAX_CHILD) {
		// Wait for atleast one child to die
		// Mentioned in the assignment
		wait(NULL);
		childNum--;
	}

	if (fork() == 0) {
	    close(socketId);
	    handleRequests(sockId);
		closeSocket(sockId);
	    exit(0);
	}
	else {
	    // Update the child count
	    childNum++;
	    close(sockId);
	    // Since we have opened it for this connection, close it now
	}
}

// readRequest: The function reads the request from the provided
// socket and returns it. In case of an error, an empty string is returned

string ProxyServer::readRequest(int sockId) {

	char getMsg[BUF_SIZE + 1];
  	int bufferSize = BUF_SIZE;

	string reqMsg = "";
	memset(getMsg, '\0', bufferSize);

	int numRead = 0;
	while((numRead = read(sockId, getMsg, bufferSize)) > 0) {

		if (numRead <= 0) {
			return "";
		}

		reqMsg += string(getMsg);
		getMsg[numRead] = '\0';
		if (DEBUGFLAG) {
			printf("Client reading %d\n", numRead);
			printf("%s", getMsg);
		}
		// if (reqMsg.find("\r\n\r\n") != string::npos)
		// 	break;
		if (reqMsg.find(CRLF + CRLF) != string::npos)
			break;
		memset(getMsg, '\0', bufferSize);
	}

	if (DEBUGFLAG) {
		cout << "The request is_____________________________\n";
		cout << reqMsg << endl;
	}

	return reqMsg;
}

// sendRequest: It takes the parsed client request, modifies it and sends
// it to the server. It returns the socket id for the remote server, which
// we then use to get the response. In case of a failed attempt it returns
// -1

int ProxyServer::sendRequest(ParsedRequest *reqParse) {
  	
	string serverRequest = "";
  	// First, create the request to send to the server
  	serverRequest += string(reqParse->method) + " " + string(reqParse->path) + \
  				 " HTTP/1.0" + CRLF;

	if (DEBUGFLAG) {
		cout << "sendRequest: The request created is " << serverRequest << endl;
	}  					 

	ParsedHeader_set(reqParse, "Connection", "close");
	ParsedHeader_set(reqParse, "Host", reqParse->host);
	// Set connection and host headers

	// Read headers
	int headSize = ParsedHeader_headersLen(reqParse);
	char* headBuffer = (char*) malloc((headSize + 10)*sizeof(char));
	if (headBuffer == NULL) {
	    cout << "sendRequest: Could not allocate enough buffer space\n";
	    return -1;
	}
 	else {
		if (DEBUGFLAG)
		    cout << "sendRequest: Now generating the server request\n";
 	}

 	// Get the parsedRequest and create the server request
 	ParsedRequest_unparse_headers(reqParse, headBuffer, headSize);
 	headBuffer[headSize] = '\0';
  	serverRequest += headBuffer;
	if (DEBUGFLAG) {
		cout << "sendRequest: The server request created is " << serverRequest << endl;
	}
	// Free the space allocated for the headers
	free(headBuffer);

  	string reqPort = "80";
  	// Default port of 80 in case port is not specified
	if (reqParse->port != NULL) {
		reqPort = reqParse->port;
	}

	// Find address of given server

	addrinfo hints;
	// The hints argument points to an addrinfo structure that specifies
	// criteria for selecting the socket address structures returned in the
	// list pointed to by res.  If hints is not NULL it points to an
	// addrinfo structure whose ai_family, ai_socktype, and ai_protocol
	// specify criteria that limit the set of socket addresses returned by
	// getaddrinfo()
	
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC; 
	hints.ai_socktype = SOCK_STREAM;
	int returnVal;
	
	// Given node and service, which identify an Internet host and a
 	// service, getaddrinfo() returns one or more addrinfo structures, each
 	// of which contains an Internet address that can be specified in a call
 	// to bind(2) or connect(2).

	addrinfo *serverList;
	// Represents a linked list contain all 
	
	if ((returnVal = getaddrinfo(reqParse->host, reqPort.c_str(), &hints, &serverList)) < 0) {
		if (DEBUGFLAG)
			cout << "sendRequest: Unable to get the address structures\n";
		return -1;
	}

	// Loop over all the results from getaddinfo and try to connect to them
	// Referred from http://man7.org/linux/man-pages/man3/getaddrinfo.3.html

	addrinfo* rp = serverList;
	int serSockId = -1;

	for (rp = serverList; rp != NULL; rp = rp->ai_next) {
	    // cout << "H" << flush;
	    if ((serSockId = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol)) < 0)
	        continue;
	    if (connect(serSockId, rp->ai_addr, rp->ai_addrlen) < 0)
	        continue;
	    break;
	}

	freeaddrinfo(serverList);

	if (rp == NULL) {
		// Always unsuccessful, thus close the connection
		closeSocket(serSockId);
	    if (DEBUGFLAG)
	    	cout << "sendRequest: Unable to requested server\n";
		return -1;
	}

	// Send request to server to check whether 
	if (write(serSockId, serverRequest.c_str(), serverRequest.size()) <= 0) {
	    return -1;
	}

	return serSockId;
}

int main(int argc, char * argv[]) {
	
	if (argc < 2) {
		cout << "Insufficient number of command line arguments\n";
		cout << "Enter port number too!\n";
		return -1;	
	}

	int portNum = atoi(argv[1]);

  	ProxyServer proxy;
	if (proxy.start(portNum) < 0) {
		cout << "main : Unable to start Proxy server\n";
	    return -1;
	}

	// Run the Proxy server
	proxy.run();

	if (DEBUGFLAG)
		cout << "Proxy server closed!\n";

	return 0;
}